/* process_monitor */
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk */
/* released under gnu gpl v2 */

/* gathers statistics on a specified process + its children */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <signal.h>

#define MAX_PIDS	2048

#define LSOF_COMMAND \
"lsof +c 0 -n -P -w -l -F pRugct"
/* lsof flags:
	-b = avoid kernel functions that might block
	+c 0 = snaplen (unlimited) for process name
	-i 4 = show ipv4 + network files
	-n = no dns
	-P = no port "dns"
	-w = suppress warnings
	-l = no login name "dns"
	-U = list unix sockets
	pRugct  = pid ppid uid gid name filetype
*/

#define PS_COMMAND \
"ps H -e -o pid,gid,ppid,comm,uid,%cpu,%mem,eip,esp,lwp,nlwp,rss,size,sz,vsz"
/* ps flags:
	H = show threads as if they were processes
	-e = select all processes
	-o ... = user defined output format
		pid = process id
		gid = effective group id
		ppid = parent pid
		comm = command name (only exe)
		uid = effective user id
		%cpu = cpu utilization of the process in "##.#" format.
               Currently, it is the CPU time used divided by the time the
               process has been running (cputime/realtime ratio),
               expressed as a percentage. It will not add up to 100%
               unless you are lucky
		%mem = ratio of the process’s resident set size  to the physical
               memory on the machine, expressed as a percentage.
		eip = instruction pointer
		esp = stack pointer
		lwp = lightweight thread id
		nlwp = # of lw threads
		rss = resident set size, the non-swapped physical memory that a
              task has used (in kiloBytes).
		size = approximate amount of swap space that would be required if
               the process were to dirty all writable pages and then be
               swapped out. This number is very rough!
		sz = size in physical pages of the core image of the process.
             This includes text, data, and stack space. Device mappings
             are currently excluded; this is subject to change.
		vsz = virtual memory size of the process in KiB (1024-byte units).
*/

struct lsof_output {
	int pid;
	int gid;
	int parent;
	char process_name[256];
	int uid;
	int num_files;
	char file_type[1024][16];
	int num_reg;
	int num_dir;
	int num_chr;
	int num_ipv4;
	int num_sock;
	int num_unix;
	int num_unknown;
};

struct ps_output {
	int pid;
	int gid;
	int parent;
	char process_name[256];
	int uid;
	double cpu;
	double mem;
	char eip[16];
	char esp[16];
	int lwp;
	int nlwp;
	int rss;
	int size;
	int sz;
	int vsz;
};

/* parameters */
static pid_t pid;  /* process id */
static unsigned long rate = 1000000;  /* rate to update logs (usecs) */
static char logfile[256] = "/var/log/process_monitor.log";  /* logfile */
static char pname[256];  /* process name */
static short debug = 0;  /* run in debug mode */
static short monitor_name = 0;  /* monitor name instead of pid */

static FILE *fp;  /* logfile fp */

char *freadline(char *line, int size, FILE *fp)
{
	char *p;

	if (fgets(line, size, fp) == NULL)
		return NULL;
	if ((p = strchr(line, '\n')) == NULL)
		return line;
	*p = '\0';

	return line;
}

/* calculate difference between two timevals */
inline unsigned long timeval_diff_usec(struct timeval *st, \
		struct timeval *et)
{
	register long long secs, usecs;

	secs = (long long) et->tv_sec - st->tv_sec;
	usecs = (long long) et->tv_usec - st->tv_usec;
	if(secs < 0)
		return 0;
	if(usecs < 0) {
		if(secs == 0)
			return 0;
		secs -= 1;
		usecs += 1000000;
	}
	return (unsigned long) secs * 1000000 + usecs;
}

static void parse_command_line(int argc, char **argv)
{
	int c;
	char *endptr[1];

	while((c = getopt(argc, argv, "dr:o:")) != -1)
		switch(c) {
			case 'd':  /* debug - log to stdout + don't background */
				debug = 1;
				break;

			case 'r':  /* rate to log at (usecs) */
				rate = atoi(optarg);
				break;

			case 'o':  /* logfile - specify full path */
				memset(logfile, 0, sizeof(logfile));
				strncpy(logfile, optarg, sizeof(logfile) - 1);
				break;

			default:
			usage:
				fprintf(stderr, \
						"usage: process_monitor [-d -r RATE -o LOGFILE] " \
						"PID|PROCESS_NAME\n" \
						"\tPID - process id of process to monitor\n" \
						"\tPROCESS_NAME - name of process to monitor\n" \
						"\t-d - debug (log to stdout + don't background)\n" \
						"\t-r RATE - rate to log at (usecs)\n" \
						"\t-o LOGFILE - logfile (default: " \
						"/var/log/process_monitor.log)\n");
				exit(EXIT_FAILURE);
				break;
		}

	if(optind != (argc - 1))
		goto usage;
	pid = (pid_t) strtol(argv[optind], endptr, 10);
	if(**endptr != '\0') {
		monitor_name = 1;
		memset(pname, 0, sizeof(pname));
		strncpy(pname, argv[optind], sizeof(pname) - 1);		
	} 
}

/* exit cleanly */
static void sigintterm_handler(int s)
{
	fclose(fp);
	_exit(EXIT_SUCCESS);
}

/* get + parse ps output into structure */
int get_ps_output(struct ps_output **psout)
{
	char line[4096];
	FILE *pfp;
	int count;
	char *str;
	
	/* get ps output */
	if(!(pfp = popen(PS_COMMAND, "r"))) {
		fprintf(stderr, "error: couldn't open pipe\n");
		_exit(EXIT_FAILURE);
	}
	count = 0;
	while(freadline(line, sizeof(line), pfp)) {
		str = strtok(line, " ");
		psout[count]->pid = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->gid = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->parent = atoi(str);
		str = strtok(NULL, " ");
		strcpy(psout[count]->process_name, str);
		str = strtok(NULL, " ");
		psout[count]->uid = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->cpu = atof(str);
		str = strtok(NULL, " ");
		psout[count]->mem = atof(str);
		str = strtok(NULL, " ");
		strcpy(psout[count]->eip, str);
		str = strtok(NULL, " ");
		strcpy(psout[count]->esp, str);
		str = strtok(NULL, " ");
		psout[count]->lwp = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->nlwp = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->rss = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->size = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->sz = atoi(str);
		str = strtok(NULL, " ");
		psout[count]->vsz = atoi(str);
		count++;
	}
	pclose(pfp);

	return count;
}

/* get + parse lsof output into structure */
int get_lsof_output(struct lsof_output **lsofout)
{
	char line[4096];
	FILE *pfp;
	int num_files, count;
	char *ret;
	
	/* get lsof output */
	if(!(pfp = popen(LSOF_COMMAND, "r"))) {
		fprintf(stderr, "error: couldn't open pipe\n");
		_exit(EXIT_FAILURE);
	}
	count = 0;
	ret = freadline(line, sizeof(line), pfp);
	while(ret) {
		lsofout[count]->pid = atoi(line + 1);
		freadline(line, sizeof(line), pfp);
		lsofout[count]->gid = atoi(line + 1);
		freadline(line, sizeof(line), pfp);
		lsofout[count]->parent = atoi(line + 1);
		freadline(line, sizeof(line), pfp);
		strcpy(lsofout[count]->process_name, line + 1);
		freadline(line, sizeof(line), pfp);
		lsofout[count]->uid = atoi(line + 1);
		num_files = lsofout[count]->num_reg = lsofout[count]->num_dir \
				= lsofout[count]->num_chr = lsofout[count]->num_ipv4 \
				= lsofout[count]->num_sock = lsofout[count]->num_unix \
				= lsofout[count]->num_unknown = 0;
		while(1) {
			if(!(ret = freadline(line, sizeof(line), pfp)))
				break;
			if(line[0] == 'p')
				break;
			strcpy(lsofout[count]->file_type[num_files], line + 1);
			if(strcmp(lsofout[count]->file_type[num_files], "REG") == 0)
				lsofout[count]->num_reg++;
			else if(strcmp(lsofout[count]->file_type[num_files], "DIR") == 0)
				lsofout[count]->num_dir++;
			else if(strcmp(lsofout[count]->file_type[num_files], "CHR") == 0)
				lsofout[count]->num_chr++;
			else if(strcmp(lsofout[count]->file_type[num_files], "IPv4") == 0)
				lsofout[count]->num_ipv4++;
			else if(strcmp(lsofout[count]->file_type[num_files], "sock") == 0)
				lsofout[count]->num_sock++;
			else if(strcmp(lsofout[count]->file_type[num_files], "unix") == 0)
				lsofout[count]->num_unix++;
			else 
				lsofout[count]->num_unknown++;
			num_files++;
		}
		lsofout[count]->num_files = num_files;
		count++;
	}
	pclose(pfp);

	return count;
}

int main(int argc, char **argv)
{
	struct ps_output **ps_process;  /* ps process list */
	struct lsof_output **lsof_process;  /* lsof process list */
	struct timeval ts, endtime;  /* timestamp */
	unsigned long flush_count, flush_loop, spenttime;
	register int i, j, lsofind;
	int ps_total_processes;  /* number of processes in list */
	int lsof_total_processes;  /* number of processes in list */
	register int processes;  /* number of monitored processes */
	pid_t pids[MAX_PIDS];  /* pids of monitored processes */
	pid_t lastpid;  /* last seen pid of monitored processes */
	int num_files, num_reg, num_dir, num_chr, num_ipv4, num_sock;
	int num_unix, num_unknown;
	double cpu, mem;
	int nlwp, rss, size, sz, vsz;
	
	parse_command_line(argc, argv);
	
	/* setup as daemon */
	if(!debug)
		if(daemon(0, 0)) {
			fprintf(stderr, "error: couldn't daemonise\n");
			exit(EXIT_FAILURE);
		}
		
	/* capture SIGINT + SIGTERM */
	if(signal(SIGINT, sigintterm_handler) == SIG_ERR || \
			signal(SIGTERM, sigintterm_handler) == SIG_ERR) {
		fprintf(stderr, "error: couldn't setup signal handling\n");
		_exit(EXIT_FAILURE);
	}

	/* open logfile */
	if(debug)
		fp = stdout;
	else
		if(!(fp = fopen(logfile, "a"))) {
			fprintf(stderr, "error: couldn't open logfile %s\n", logfile);
			_exit(EXIT_FAILURE);
		}

	/* allocate memory for output */
	if(!(ps_process = calloc(MAX_PIDS, sizeof(struct ps_output *)))) {
		fprintf(stderr, "error: couldn't allocate memory for output\n");
		_exit(EXIT_FAILURE);
	}
	for(i = 0; i < MAX_PIDS; i++)
		if(!(ps_process[i] = malloc(sizeof(struct ps_output)))) {
			fprintf(stderr, "error: couldn't allocate memory for output\n");
			_exit(EXIT_FAILURE);
		}
	if(!(lsof_process = calloc(MAX_PIDS, sizeof(struct lsof_output *)))) {
		fprintf(stderr, "error: couldn't allocate memory for output\n");
		_exit(EXIT_FAILURE);
	}
	for(i = 0; i < MAX_PIDS; i++)
		if(!(lsof_process[i] = malloc(sizeof(struct lsof_output)))) {
			fprintf(stderr, "error: couldn't allocate memory for output\n");
			_exit(EXIT_FAILURE);
		}
		
	/* periodically log data */
	flush_count = flush_loop = 60000000 / rate;
	lastpid = 0;
	while(1) {
		/* timestamp */
		gettimeofday(&ts, NULL);

		/* get ps output */
		ps_total_processes = get_ps_output(ps_process);
		
		/* get lsof output */
		lsof_total_processes = get_lsof_output(lsof_process);

		/* extract data from lsof process snapshot */
		processes = 0;
		if(monitor_name) {
			/* monitor by name, so find pid with specified name */
			for(i = 0; i < ps_total_processes; i++)
				if(strcmp(ps_process[i]->process_name, pname) == 0) {
					/* it's the monitored process - save pid */
					processes++;
					break;
				}
		} else {
			/* monitor by pid, so find process with specified pid */
			for(i = 0; i < ps_total_processes; i++)
				if(ps_process[i]->pid == pid) {
					/* it's the monitored process - save name */
					strncpy(pname, ps_process[i]->process_name, \
							sizeof(pname) - 1);
					processes++;
					break;
				}
		}
		if(!processes && monitor_name && lastpid) {
			 /* maybe changed name */
			for(i = 0; i < ps_total_processes; i++)
				if(ps_process[i]->pid == lastpid) {
					/* changed name */
					strncpy(pname, ps_process[i]->process_name, \
							sizeof(pname) - 1);
					processes++;
					break;
				}
			if(!processes)
				lastpid = 0;
		}
		if(!processes) {
			/* no processes with that name or pid */
			fprintf(fp, "%d.%06d %s 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0 0\n", \
					(int) ts.tv_sec, (int) ts.tv_usec, \
					monitor_name ? pname : "(killed)");
			goto done;
		}
		for(lsofind = 0; lsofind < lsof_total_processes; lsofind++)
			if(lsof_process[lsofind]->pid == ps_process[i]->pid)
				break;
		
		/* a process - so calculate stats for itself */
		pids[0] = lastpid = ps_process[i]->pid;
		cpu = ps_process[i]->cpu;
		mem = ps_process[i]->mem;
		nlwp = ps_process[i]->nlwp;
		rss = ps_process[i]->rss;
		size = ps_process[i]->size;
		sz = ps_process[i]->sz;
		vsz = ps_process[i]->vsz;
		if(lsofind < lsof_total_processes) {
			num_files = lsof_process[lsofind]->num_files;
			num_reg = lsof_process[lsofind]->num_reg;
			num_dir = lsof_process[lsofind]->num_dir;
			num_chr = lsof_process[lsofind]->num_chr;
			num_ipv4 = lsof_process[lsofind]->num_ipv4;
			num_sock = lsof_process[lsofind]->num_sock;
			num_unix = lsof_process[lsofind]->num_unix;
			num_unknown = lsof_process[lsofind]->num_unknown;
		} else
			num_files = num_reg = num_dir = num_chr = num_ipv4 = \
					num_sock = num_unix = num_unknown = 0;

		/* + stats for children */
		for(i = 0; i < ps_total_processes; i++)
			for(j = 0; j < processes; j++)
				if(ps_process[i]->parent == pids[j] && processes < MAX_PIDS) {
						/* it's a child of the monitored process */
						for(lsofind = 0; lsofind < lsof_total_processes; \
								lsofind++)
							if(lsof_process[lsofind]->pid == ps_process[i]->pid)
								break;
						pids[processes++] = ps_process[i]->pid;
						cpu += ps_process[i]->cpu;
						mem += ps_process[i]->mem;
						nlwp += ps_process[i]->nlwp;
						rss += ps_process[i]->rss;
						size += ps_process[i]->size;
						sz += ps_process[i]->sz;
						vsz += ps_process[i]->vsz;
						if(lsofind < lsof_total_processes) {
							num_files += lsof_process[lsofind]->num_files;
							num_reg += lsof_process[lsofind]->num_reg;
							num_dir += lsof_process[lsofind]->num_dir;
							num_chr += lsof_process[lsofind]->num_chr;
							num_ipv4 += lsof_process[lsofind]->num_ipv4;
							num_sock += lsof_process[lsofind]->num_sock;
							num_unix += lsof_process[lsofind]->num_unix;
							num_unknown += lsof_process[lsofind]->num_unknown;
						}
						break;
				}

		/* output to logfile */
		fprintf(fp, \
			"%d.%06d %s %d %d %f %f %d %d %d %d %d %d %d %d %d %d %d %d %d\n", \
			(int) ts.tv_sec, (int) ts.tv_usec, \
			pname, pids[0], processes, \
			cpu, mem, nlwp, rss, size, sz, vsz, \
			num_files, num_reg, num_dir, num_chr, num_ipv4, \
			num_sock, num_unix, num_unknown);
	done:		
		/* flush logfile every 60 secs */
		if(debug || !flush_count--) {
			flush_count = flush_loop;
			fflush(fp);
		}
							
		/* sleep */
		gettimeofday(&endtime, NULL);
		spenttime = timeval_diff_usec(&ts, &endtime);
		if(spenttime < rate)
			usleep(rate - spenttime);
	}
}
